/*
 * Copyright notice from original mutt:
 * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
 * Copyright (C) 1999-2002 Thomas Roessler <roessler@does-not-exist.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

/*
 * This file contains routines specific to MH and ``maildir'' style
 * mailboxes.
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include "mutt.h"
#include "mx.h"
#include "mh.h"
#include "mbox.h"
#include "copy.h"
#include "buffy.h"
#include "sort.h"
#include "thread.h"
#include "hcache.h"

#include "lib/mem.h"
#include "lib/intl.h"
#include "lib/str.h"
#include "lib/debug.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <utime.h>

#if HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

struct maildir {
  HEADER *h;
  char *canon_fname;
  unsigned header_parsed:1;
#ifdef USE_INODESORT
  ino_t inode;
#endif /* USE_INODESORT */
  struct maildir *next;
};

struct mh_sequences {
  int max;
  short *flags;
};

/* mh_sequences support */

#define MH_SEQ_UNSEEN  (1 << 0)
#define MH_SEQ_REPLIED (1 << 1)
#define MH_SEQ_FLAGGED (1 << 2)

/* prototypes */
static int maildir_check_empty (const char*);
static int maildir_check_mailbox (CONTEXT*, int*, int);
static int mh_check_mailbox (CONTEXT*, int*, int);

static void mhs_alloc (struct mh_sequences *mhs, int i)
{
  int j;
  int newmax;

  if (i > mhs->max || !mhs->flags) {
    newmax = i + 128;
    mem_realloc (&mhs->flags, sizeof (mhs->flags[0]) * (newmax + 1));
    for (j = mhs->max + 1; j <= newmax; j++)
      mhs->flags[j] = 0;

    mhs->max = newmax;
  }
}

static void mhs_free_sequences (struct mh_sequences *mhs)
{
  mem_free (&mhs->flags);
}

static short mhs_check (struct mh_sequences *mhs, int i)
{
  if (!mhs->flags || i > mhs->max)
    return 0;
  else
    return mhs->flags[i];
}

static short mhs_set (struct mh_sequences *mhs, int i, short f)
{
  mhs_alloc (mhs, i);
  mhs->flags[i] |= f;
  return mhs->flags[i];
}

#if 0

/* unused */

static short mhs_unset (struct mh_sequences *mhs, int i, short f)
{
  mhs_alloc (mhs, i);
  mhs->flags[i] &= ~f;
  return mhs->flags[i];
}

#endif

static void mh_read_token (char *t, int *first, int *last)
{
  char *p;

  if ((p = strchr (t, '-'))) {
    *p++ = '\0';
    *first = atoi (t);
    *last = atoi (p);
  }
  else
    *first = *last = atoi (t);
}

static void mh_read_sequences (struct mh_sequences *mhs, const char *path)
{
  FILE *fp;
  int line = 1;
  char *buff = NULL;
  char *t;
  size_t sz = 0;

  short f;
  int first, last;

  char pathname[_POSIX_PATH_MAX];

  snprintf (pathname, sizeof (pathname), "%s/.mh_sequences", path);

  if (!(fp = fopen (pathname, "r")))
    return;

  while ((buff = mutt_read_line (buff, &sz, fp, &line))) {
    if (!(t = strtok (buff, " \t:")))
      continue;

    if (!str_cmp (t, MhUnseen))
      f = MH_SEQ_UNSEEN;
    else if (!str_cmp (t, MhFlagged))
      f = MH_SEQ_FLAGGED;
    else if (!str_cmp (t, MhReplied))
      f = MH_SEQ_REPLIED;
    else                        /* unknown sequence */
      continue;

    while ((t = strtok (NULL, " \t:"))) {
      mh_read_token (t, &first, &last);
      for (; first <= last; first++)
        mhs_set (mhs, first, f);
    }
  }

  mem_free (&buff);
  safe_fclose (&fp);
}

int mh_buffy (const char *path)
{
  int i, r = 0;
  struct mh_sequences mhs;

  memset (&mhs, 0, sizeof (mhs));

  mh_read_sequences (&mhs, path);
  for (i = 0; !r && i <= mhs.max; i++)
    if (mhs_check (&mhs, i) & MH_SEQ_UNSEEN)
      r = 1;
  mhs_free_sequences (&mhs);
  return r;
}

static int mh_mkstemp (CONTEXT * dest, FILE ** fp, char **tgt)
{
  int fd;
  char path[_POSIX_PATH_MAX];

  FOREVER {
    snprintf (path, _POSIX_PATH_MAX, "%s/.mutt-%s-%d-%d",
              dest->path, NONULL (Hostname), (int) getpid (), Counter++);
    umask (Umask);
    if ((fd = open (path, O_WRONLY | O_EXCL | O_CREAT, 0666)) == -1) {
      if (errno != EEXIST) {
        mutt_perror (path);
        return -1;
      }
    }
    else {
      *tgt = str_dup (path);
      break;
    }
  }

  if ((*fp = fdopen (fd, "w")) == NULL) {
    mem_free (tgt);
    close (fd);
    unlink (path);
    return (-1);
  }

  return 0;
}

static void mhs_write_one_sequence (FILE * fp, struct mh_sequences *mhs,
                                    short f, const char *tag)
{
  int i;
  int first, last;

  fprintf (fp, "%s:", tag);

  first = -1;
  last = -1;

  for (i = 0; i <= mhs->max; i++) {
    if ((mhs_check (mhs, i) & f)) {
      if (first < 0)
        first = i;
      else
        last = i;
    }
    else if (first >= 0) {
      if (last < 0)
        fprintf (fp, " %d", first);
      else
        fprintf (fp, " %d-%d", first, last);

      first = -1;
      last = -1;
    }
  }

  if (first >= 0) {
    if (last < 0)
      fprintf (fp, " %d", first);
    else
      fprintf (fp, " %d-%d", first, last);
  }

  fputc ('\n', fp);
}

/* XXX - we don't currently remove deleted messages from sequences we don't know.  Should we? */

void mh_update_sequences (CONTEXT * ctx)
{
  FILE *ofp, *nfp;

  char sequences[_POSIX_PATH_MAX];
  char *tmpfname;
  char *buff = NULL;
  char *p;
  size_t s;
  int l = 0;
  int i;

  int unseen = 0;
  int flagged = 0;
  int replied = 0;

  char seq_unseen[STRING];
  char seq_replied[STRING];
  char seq_flagged[STRING];


  struct mh_sequences mhs;

  memset (&mhs, 0, sizeof (mhs));

  snprintf (seq_unseen, sizeof (seq_unseen), "%s:", NONULL (MhUnseen));
  snprintf (seq_replied, sizeof (seq_replied), "%s:", NONULL (MhReplied));
  snprintf (seq_flagged, sizeof (seq_flagged), "%s:", NONULL (MhFlagged));

  if (mh_mkstemp (ctx, &nfp, &tmpfname) != 0) {
    /* error message? */
    return;
  }

  snprintf (sequences, sizeof (sequences), "%s/.mh_sequences", ctx->path);


  /* first, copy unknown sequences */
  if ((ofp = fopen (sequences, "r"))) {
    while ((buff = mutt_read_line (buff, &s, ofp, &l))) {
      if (!str_ncmp (buff, seq_unseen, str_len (seq_unseen)))
        continue;
      if (!str_ncmp (buff, seq_flagged, str_len (seq_flagged)))
        continue;
      if (!str_ncmp (buff, seq_replied, str_len (seq_replied)))
        continue;

      fprintf (nfp, "%s\n", buff);
    }
  }
  safe_fclose (&ofp);

  /* now, update our unseen, flagged, and replied sequences */
  for (l = 0; l < ctx->msgcount; l++) {
    if (ctx->hdrs[l]->deleted)
      continue;

    if ((p = strrchr (ctx->hdrs[l]->path, '/')))
      p++;
    else
      p = ctx->hdrs[l]->path;

    i = atoi (p);

    if (!ctx->hdrs[l]->read) {
      mhs_set (&mhs, i, MH_SEQ_UNSEEN);
      unseen++;
    }
    if (ctx->hdrs[l]->flagged) {
      mhs_set (&mhs, i, MH_SEQ_FLAGGED);
      flagged++;
    }
    if (ctx->hdrs[l]->replied) {
      mhs_set (&mhs, i, MH_SEQ_REPLIED);
      replied++;
    }
  }

  /* write out the new sequences */
  if (unseen)
    mhs_write_one_sequence (nfp, &mhs, MH_SEQ_UNSEEN, NONULL (MhUnseen));
  if (flagged)
    mhs_write_one_sequence (nfp, &mhs, MH_SEQ_FLAGGED, NONULL (MhFlagged));
  if (replied)
    mhs_write_one_sequence (nfp, &mhs, MH_SEQ_REPLIED, NONULL (MhReplied));

  mhs_free_sequences (&mhs);


  /* try to commit the changes - no guarantee here */
  safe_fclose (&nfp);

  unlink (sequences);
  if (safe_rename (tmpfname, sequences) != 0) {
    /* report an error? */
    unlink (tmpfname);
  }

  mem_free (&tmpfname);
}

static void mh_sequences_add_one (CONTEXT * ctx, int n, short unseen,
                                  short flagged, short replied)
{
  short unseen_done = 0;
  short flagged_done = 0;
  short replied_done = 0;

  FILE *ofp = NULL, *nfp = NULL;

  char *tmpfname;
  char sequences[_POSIX_PATH_MAX];

  char seq_unseen[STRING];
  char seq_replied[STRING];
  char seq_flagged[STRING];

  char *buff = NULL;
  int line;
  size_t sz;

  if (mh_mkstemp (ctx, &nfp, &tmpfname) == -1)
    return;

  snprintf (seq_unseen, sizeof (seq_unseen), "%s:", NONULL (MhUnseen));
  snprintf (seq_replied, sizeof (seq_replied), "%s:", NONULL (MhReplied));
  snprintf (seq_flagged, sizeof (seq_flagged), "%s:", NONULL (MhFlagged));

  snprintf (sequences, sizeof (sequences), "%s/.mh_sequences", ctx->path);
  if ((ofp = fopen (sequences, "r"))) {
    while ((buff = mutt_read_line (buff, &sz, ofp, &line))) {
      if (unseen && !strncmp (buff, seq_unseen, str_len (seq_unseen))) {
        fprintf (nfp, "%s %d\n", buff, n);
        unseen_done = 1;
      }
      else if (flagged
               && !strncmp (buff, seq_flagged, str_len (seq_flagged))) {
        fprintf (nfp, "%s %d\n", buff, n);
        flagged_done = 1;
      }
      else if (replied
               && !strncmp (buff, seq_replied, str_len (seq_replied))) {
        fprintf (nfp, "%s %d\n", buff, n);
        replied_done = 1;
      }
      else
        fprintf (nfp, "%s\n", buff);
    }
  }
  safe_fclose (&ofp);
  mem_free (&buff);

  if (!unseen_done && unseen)
    fprintf (nfp, "%s: %d\n", NONULL (MhUnseen), n);
  if (!flagged_done && flagged)
    fprintf (nfp, "%s: %d\n", NONULL (MhFlagged), n);
  if (!replied_done && replied)
    fprintf (nfp, "%s: %d\n", NONULL (MhReplied), n);

  safe_fclose (&nfp);

  unlink (sequences);
  if (safe_rename (tmpfname, sequences) != 0)
    unlink (tmpfname);

  mem_free (&tmpfname);
}

static void mh_update_maildir (struct maildir *md, struct mh_sequences *mhs)
{
  int i;
  short f;
  char *p;

  for (; md; md = md->next) {
    if ((p = strrchr (md->h->path, '/')))
      p++;
    else
      p = md->h->path;

    i = atoi (p);
    f = mhs_check (mhs, i);

    md->h->read = (f & MH_SEQ_UNSEEN) ? 0 : 1;
    md->h->flagged = (f & MH_SEQ_FLAGGED) ? 1 : 0;
    md->h->replied = (f & MH_SEQ_REPLIED) ? 1 : 0;
  }
}

/* maildir support */

static void maildir_free_entry (struct maildir **md)
{
  if (!md || !*md)
    return;

  mem_free (&(*md)->canon_fname);
  if ((*md)->h)
    mutt_free_header (&(*md)->h);

  mem_free (md);
}

static void maildir_free_maildir (struct maildir **md)
{
  struct maildir *p, *q;

  if (!md || !*md)
    return;

  for (p = *md; p; p = q) {
    q = p->next;
    maildir_free_entry (&p);
  }
}

static void maildir_parse_flags (HEADER * h, const char *path)
{
  char *p, *q = NULL;

  h->flagged = 0;
  h->read = 0;
  h->replied = 0;

  if ((p = strrchr (path, ':')) != NULL && str_ncmp (p + 1, "2,", 2) == 0) {
    p += 3;

    str_replace (&h->maildir_flags, p);
    q = h->maildir_flags;

    while (*p) {
      switch (*p) {
      case 'F':

        h->flagged = 1;
        break;

      case 'S':                /* seen */

        h->read = 1;
        break;

      case 'R':                /* replied */

        h->replied = 1;
        break;

      case 'T':                /* trashed */
        h->trash = 1;
        h->deleted = 1;
        break;

      default:
        *q++ = *p;
        break;
      }
      p++;
    }
  }

  if (q == h->maildir_flags)
    mem_free (&h->maildir_flags);
  else if (q)
    *q = '\0';
}

static void maildir_update_mtime (CONTEXT * ctx)
{
  char buf[_POSIX_PATH_MAX];
  struct stat st;

  if (ctx->magic == M_MAILDIR) {
    snprintf (buf, sizeof (buf), "%s/%s", ctx->path, "cur");
    if (stat (buf, &st) == 0)
      ctx->mtime_cur = st.st_mtime;
    snprintf (buf, sizeof (buf), "%s/%s", ctx->path, "new");
  }
  else {
    snprintf (buf, sizeof (buf), "%s/.mh_sequences", ctx->path);
    if (stat (buf, &st) == 0)
      ctx->mtime_cur = st.st_mtime;

    strfcpy (buf, ctx->path, sizeof (buf));
  }

  if (stat (buf, &st) == 0)
    ctx->mtime = st.st_mtime;
}

/* 
 * Actually parse a maildir message.  This may also be used to fill
 * out a fake header structure generated by lazy maildir parsing.
 */
static HEADER *maildir_parse_message (int magic, const char *fname,
                                      int is_old, HEADER * _h)
{
  FILE *f;
  HEADER *h = _h;
  struct stat st;

  if ((f = fopen (fname, "r")) != NULL) {
    if (!h)
      h = mutt_new_header ();
    h->env = mutt_read_rfc822_header (f, h, 0, 0);

    fstat (fileno (f), &st);
    fclose (f);

    if (!h->received)
      h->received = h->date_sent;

    if (h->content->length <= 0)
      h->content->length = st.st_size - h->content->offset;

    h->index = -1;

    if (magic == M_MAILDIR) {
      /* 
       * maildir stores its flags in the filename, so ignore the
       * flags in the header of the message 
       */

      h->old = is_old;
      maildir_parse_flags (h, fname);
    }
    return h;
  }
  return NULL;
}

/* 
 * Note that this routine will _not_ modify the context given by
 * ctx. 
 *
 * It's used in the first parsing pass on maildir and MH folders.
 * In the MH case, this means full parsing of the folder.  In the
 * maildir case, it means that we only look at flags, and create a
 * fake HEADER structure, which may later be filled in by
 * maildir_parse_message(), when called from
 * maildir_delayed_parsing().
 * 
 */

static int maildir_parse_entry (CONTEXT * ctx, struct maildir ***last,
                                const char *subdir, const char *fname,
                                int *count, int is_old, ino_t inode)
{
  struct maildir *entry;
  HEADER *h = NULL;
  char buf[_POSIX_PATH_MAX];

  if (subdir)
    snprintf (buf, sizeof (buf), "%s/%s/%s", ctx->path, subdir, fname);
  else
    snprintf (buf, sizeof (buf), "%s/%s", ctx->path, fname);

  if (ctx->magic == M_MH)
    h = maildir_parse_message (ctx->magic, buf, is_old, NULL);
  else {
    h = mutt_new_header ();
    h->old = is_old;
    maildir_parse_flags (h, buf);
  }

  if (h != NULL) {
    if (count) {
      (*count)++;
      if (!ctx->quiet && ReadInc && ((*count % ReadInc) == 0 || *count == 1))
        mutt_message (_("Reading %s... %d"), ctx->path, *count);
    }

    if (subdir) {
      snprintf (buf, sizeof (buf), "%s/%s", subdir, fname);
      h->path = str_dup (buf);
    }
    else
      h->path = str_dup (fname);

    entry = mem_calloc (sizeof (struct maildir), 1);
    entry->h = h;
    entry->header_parsed = (ctx->magic == M_MH);
#ifdef USE_INODESORT
    entry->inode = inode;
#endif /* USE_INODESORT */
    **last = entry;
    *last = &entry->next;

    return 0;
  }

  return -1;
}



/* Ignore the garbage files.  A valid MH message consists of only
 * digits.  Deleted message get moved to a filename with a comma before
 * it.
 */

int mh_valid_message (const char *s)
{
  for (; *s; s++) {
    if (!isdigit ((unsigned char) *s))
      return 0;
  }
  return 1;
}

static int maildir_parse_dir (CONTEXT * ctx, struct maildir ***last,
                              const char *subdir, int *count)
{
  DIR *dirp;
  struct dirent *de;
  char buf[_POSIX_PATH_MAX];
  int is_old = 0;

  if (subdir) {
    snprintf (buf, sizeof (buf), "%s/%s", ctx->path, subdir);
    is_old = (str_cmp ("cur", subdir) == 0);
  }
  else
    strfcpy (buf, ctx->path, sizeof (buf));

  if ((dirp = opendir (buf)) == NULL)
    return -1;

  while ((de = readdir (dirp)) != NULL) {

    if ((ctx->magic == M_MH && !mh_valid_message (de->d_name))
        || (ctx->magic == M_MAILDIR && *de->d_name == '.'))
      continue;

    /* FOO - really ignore the return value? */

    debug_print (2, ("parsing %s\n", de->d_name));
    maildir_parse_entry (ctx, last, subdir, de->d_name, count, is_old,
                         de->d_ino);
  }

  closedir (dirp);
  return 0;
}

static int maildir_add_to_context (CONTEXT * ctx, struct maildir *md)
{
  int oldmsgcount = ctx->msgcount;

  while (md) {

    debug_print (2, ("considering %s\n", NONULL (md->canon_fname)));

    if (md->h) {
      debug_print (2, ("flags: %s%s%s%s%s\n", md->h->flagged ? "f" : "",
                  md->h->deleted ? "D" : "", md->h->replied ? "r" : "",
                  md->h->old ? "O" : "", md->h->read ? "R" : ""));
      if (ctx->msgcount == ctx->hdrmax)
        mx_alloc_memory (ctx);

      ctx->hdrs[ctx->msgcount] = md->h;
      ctx->hdrs[ctx->msgcount]->index = ctx->msgcount;
      ctx->size +=
        md->h->content->length + md->h->content->offset -
        md->h->content->hdr_offset;

      md->h = NULL;
      ctx->msgcount++;
    }
    md = md->next;
  }

  if (ctx->msgcount > oldmsgcount) {
    mx_update_context (ctx, ctx->msgcount - oldmsgcount);
    return 1;
  }
  return 0;
}

static int maildir_move_to_context (CONTEXT * ctx, struct maildir **md)
{
  int r;

  r = maildir_add_to_context (ctx, *md);
  maildir_free_maildir (md);
  return r;
}

#ifdef USE_INODESORT
/*
 * Merge two maildir lists according to the inode numbers.
 */
static struct maildir *maildir_merge_inode (struct maildir *left,
                                            struct maildir *right)
{
  struct maildir *head;
  struct maildir *tail;

  if (left && right) {
    if (left->inode < right->inode) {
      head = left;
      left = left->next;
    }
    else {
      head = right;
      right = right->next;
    }
  }
  else {
    if (left)
      return left;
    else
      return right;
  }

  tail = head;

  while (left && right) {
    if (left->inode < right->inode) {
      tail->next = left;
      left = left->next;
    }
    else {
      tail->next = right;
      right = right->next;
    }
    tail = tail->next;
  }

  if (left) {
    tail->next = left;
  }
  else {
    tail->next = right;
  }

  return head;
}

/*
 * Sort maildir list according to inode.
 */
static struct maildir *maildir_sort_inode (struct maildir *list)
{
  struct maildir *left = list;
  struct maildir *right = list;

  if (!list || !list->next) {
    return list;
  }

  list = list->next;
  while (list && list->next) {
    right = right->next;
    list = list->next->next;
  }

  list = right;
  right = right->next;
  list->next = 0;

  left = maildir_sort_inode (left);
  right = maildir_sort_inode (right);
  return maildir_merge_inode (left, right);
}
#endif /* USE_INODESORT */

#if USE_HCACHE
static size_t maildir_hcache_keylen (const char *fn)
{
  const char *p = strchr (fn, ':');

  return p ? (size_t) (p - fn) : str_len (fn);
}
#endif

/* 
 * This function does the second parsing pass for a maildir-style
 * folder.
 */
void maildir_delayed_parsing (CONTEXT * ctx, struct maildir *md)
{
  struct maildir *p;
  char fn[_POSIX_PATH_MAX];
  int count;

#if USE_HCACHE
  void *hc = NULL;
  void *data;
  struct timeval *when = NULL;
  struct stat lastchanged;
  int ret;

  hc = mutt_hcache_open (HeaderCache, ctx->path);
#endif

  for (p = md, count = 0; p; p = p->next, count++) {
    if (!(p && p->h && !p->header_parsed))
      continue;

#if USE_HCACHE
    data = mutt_hcache_fetch (hc, p->h->path + 3, &maildir_hcache_keylen);
    when = (struct timeval *) data;
#endif

    if (!ctx->quiet && ReadInc && ((count % ReadInc) == 0 || count == 1))
      mutt_message (_("Reading %s... %d"), ctx->path, count);
    snprintf (fn, sizeof (fn), "%s/%s", ctx->path, p->h->path);

#if USE_HCACHE
    if (option (OPTHCACHEVERIFY)) {
      ret = stat (fn, &lastchanged);
    }
    else {
      lastchanged.st_mtime = 0;
      ret = 0;
    }

    if (data != NULL && !ret && lastchanged.st_mtime <= when->tv_sec) {
      p->h = mutt_hcache_restore ((unsigned char *) data, &p->h);
      maildir_parse_flags (p->h, fn);
    }
    else
#endif
    if (maildir_parse_message (ctx->magic, fn, p->h->old, p->h)) {
      p->header_parsed = 1;
      maildir_parse_flags (p->h, fn);
#if USE_HCACHE
      mutt_hcache_store (hc, p->h->path + 3, p->h, 0, &maildir_hcache_keylen);
#endif
    }
    else
      mutt_free_header (&p->h);
#if USE_HCACHE
    mem_free (&data);
#endif
  }
#if USE_HCACHE
  mutt_hcache_close (hc);
#endif
}

/* Read a MH/maildir style mailbox.
 *
 * args:
 *	ctx [IN/OUT]	context for this mailbox
 *	subdir [IN]	NULL for MH mailboxes, otherwise the subdir of the
 *			maildir mailbox to read from
 */
static int _mh_read_dir (CONTEXT * ctx, const char *subdir)
{
  struct maildir *md;
  struct mh_sequences mhs;
  struct maildir **last;
  int count;


  memset (&mhs, 0, sizeof (mhs));

  maildir_update_mtime (ctx);

  md = NULL;
  last = &md;
  count = 0;
  if (maildir_parse_dir (ctx, &last, subdir, &count) == -1)
    return -1;

  if (ctx->magic == M_MH) {
    mh_read_sequences (&mhs, ctx->path);
    mh_update_maildir (md, &mhs);
    mhs_free_sequences (&mhs);
  }

#ifdef USE_INODESORT
  md = maildir_sort_inode (md);
#endif /* USE_INODESORT */

  if (ctx->magic == M_MAILDIR)
    maildir_delayed_parsing (ctx, md);

  maildir_move_to_context (ctx, &md);
  return 0;
}

static int mh_read_dir (CONTEXT* ctx) {
  return (_mh_read_dir (ctx, NULL));
}

/* read a maildir style mailbox */
static int maildir_read_dir (CONTEXT * ctx)
{
  /* maildir looks sort of like MH, except that there are two subdirectories
   * of the main folder path from which to read messages
   */
  if (_mh_read_dir (ctx, "new") == -1 || _mh_read_dir (ctx, "cur") == -1)
    return (-1);

  return 0;
}

/*
 * Open a new (temporary) message in an MH folder.
 */

static int mh_open_new_message (MESSAGE * msg, CONTEXT * dest, HEADER * hdr)
{
  return mh_mkstemp (dest, &msg->fp, &msg->path);
}

int ch_compar (const void *a, const void *b)
{
  return (int) (*((const char *) a) - *((const char *) b));
}

static void maildir_flags (char *dest, size_t destlen, HEADER * hdr)
{
  *dest = '\0';

  /*
   * The maildir specification requires that all files in the cur
   * subdirectory have the :unique string appeneded, regardless of whether
   * or not there are any flags.  If .old is set, we know that this message
   * will end up in the cur directory, so we include it in the following
   * test even though there is no associated flag.
   */

  if (hdr
      && (hdr->flagged || hdr->replied || hdr->read || hdr->deleted
          || hdr->old || hdr->maildir_flags)) {
    char tmp[LONG_STRING];

    snprintf (tmp, sizeof (tmp),
              "%s%s%s%s%s",
              hdr->flagged ? "F" : "",
              hdr->replied ? "R" : "",
              hdr->read ? "S" : "", hdr->deleted ? "T" : "",
              NONULL (hdr->maildir_flags));
    if (hdr->maildir_flags)
      qsort (tmp, str_len (tmp), 1, ch_compar);
    snprintf (dest, destlen, ":2,%s", tmp);
  }
}


/*
 * Open a new (temporary) message in a maildir folder.
 * 
 * Note that this uses _almost_ the maildir file name format, but
 * with a {cur,new} prefix.
 *
 */

static int maildir_open_new_message (MESSAGE * msg, CONTEXT * dest, HEADER * hdr)
{
  int fd;
  char path[_POSIX_PATH_MAX];
  char suffix[16];
  char subdir[16];

  if (hdr) {
    short deleted = hdr->deleted;

    hdr->deleted = 0;

    maildir_flags (suffix, sizeof (suffix), hdr);

    hdr->deleted = deleted;
  }
  else
    *suffix = '\0';

  if (hdr && (hdr->read || hdr->old))
    strfcpy (subdir, "cur", sizeof (subdir));
  else
    strfcpy (subdir, "new", sizeof (subdir));

  FOREVER {
    snprintf (path, _POSIX_PATH_MAX, "%s/tmp/%s.%ld.%u_%d.%s%s",
              dest->path, subdir, (long) time (NULL),
              (unsigned int) getpid (), Counter++, NONULL (Hostname), suffix);

    debug_print (2, ("trying %s.\n", path));

    umask (Umask);
    if ((fd = open (path, O_WRONLY | O_EXCL | O_CREAT, 0666)) == -1) {
      if (errno != EEXIST) {
        mutt_perror (path);
        return -1;
      }
    }
    else {
      debug_print (2, ("success.\n"));
      msg->path = str_dup (path);
      break;
    }
  }

  if ((msg->fp = fdopen (fd, "w")) == NULL) {
    mem_free (&msg->path);
    close (fd);
    unlink (path);
    return (-1);
  }

  return 0;
}



/*
 * Commit a message to a maildir folder.
 * 
 * msg->path contains the file name of a file in tmp/. We take the
 * flags from this file's name. 
 *
 * ctx is the mail folder we commit to.
 * 
 * hdr is a header structure to which we write the message's new
 * file name.  This is used in the mh and maildir folder synch
 * routines.  When this routine is invoked from mx_commit_message,
 * hdr is NULL. 
 *
 * msg->path looks like this:
 * 
 *    tmp/{cur,new}.mutt-HOSTNAME-PID-COUNTER:flags
 * 
 * See also maildir_open_new_message().
 * 
 */

static int maildir_commit_message (MESSAGE * msg, CONTEXT * ctx, HEADER * hdr)
{
  char subdir[4];
  char suffix[16];
  char path[_POSIX_PATH_MAX];
  char full[_POSIX_PATH_MAX];
  char *s;

  if (safe_fclose (&msg->fp) != 0)
    return -1;

  /* extract the subdir */
  s = strrchr (msg->path, '/') + 1;
  strfcpy (subdir, s, 4);

  /* extract the flags */
  if ((s = strchr (s, ':')))
    strfcpy (suffix, s, sizeof (suffix));
  else
    suffix[0] = '\0';

  /* construct a new file name. */
  FOREVER {
    snprintf (path, _POSIX_PATH_MAX, "%s/%ld.%u_%d.%s%s", subdir,
              (long) time (NULL), (unsigned int) getpid (), Counter++,
              NONULL (Hostname), suffix);
    snprintf (full, _POSIX_PATH_MAX, "%s/%s", ctx->path, path);

    debug_print (2, ("renaming %s to %s.\n", msg->path, full));

    if (safe_rename (msg->path, full) == 0) {
      if (hdr)
        str_replace (&hdr->path, path);
      mem_free (&msg->path);

      /*
       * Adjust the mtime on the file to match the time at which this
       * message was received.  Currently this is only set when copying
       * messages between mailboxes, so we test to ensure that it is
       * actually set.
       */
      if (msg->received) {
        struct utimbuf ut;

        ut.actime = msg->received;
        ut.modtime = msg->received;
        if (utime (full, &ut)) {
          mutt_perror (_
                       ("maildir_commit_message(): unable to set time on file"));
          return -1;
        }
      }

      return 0;
    }
    else if (errno != EEXIST) {
      mutt_perror (ctx->path);
      return -1;
    }
  }
}

/* 
 * commit a message to an MH folder.
 * 
 */


static int _mh_commit_message (MESSAGE * msg, CONTEXT * ctx, HEADER * hdr,
                               short updseq)
{
  DIR *dirp;
  struct dirent *de;
  char *cp, *dep;
  unsigned int n, hi = 0;
  char path[_POSIX_PATH_MAX];
  char tmp[16];

  if (safe_fclose (&msg->fp) != 0)
    return -1;

  if ((dirp = opendir (ctx->path)) == NULL) {
    mutt_perror (ctx->path);
    return (-1);
  }

  /* figure out what the next message number is */
  while ((de = readdir (dirp)) != NULL) {
    dep = de->d_name;
    if (*dep == ',')
      dep++;
    cp = dep;
    while (*cp) {
      if (!isdigit ((unsigned char) *cp))
        break;
      cp++;
    }
    if (!*cp) {
      n = atoi (dep);
      if (n > hi)
        hi = n;
    }
  }
  closedir (dirp);

  /* 
   * Now try to rename the file to the proper name.
   * 
   * Note: We may have to try multiple times, until we find a free
   * slot.
   */

  FOREVER {
    hi++;
    snprintf (tmp, sizeof (tmp), "%d", hi);
    snprintf (path, sizeof (path), "%s/%s", ctx->path, tmp);
    if (safe_rename (msg->path, path) == 0) {
      if (hdr)
        str_replace (&hdr->path, tmp);
      mem_free (&msg->path);
      break;
    }
    else if (errno != EEXIST) {
      mutt_perror (ctx->path);
      return -1;
    }
  }
  if (updseq)
    mh_sequences_add_one (ctx, hi, !msg->flags.read, msg->flags.flagged,
                          msg->flags.replied);
  return 0;
}

static int mh_commit_message (MESSAGE * msg, CONTEXT * ctx, HEADER * hdr) {
  return _mh_commit_message (msg, ctx, hdr, 1);
}

/* Sync a message in an MH folder.
 * 
 * This code is also used for attachment deletion in maildir
 * folders.
 */

static int mh_rewrite_message (CONTEXT * ctx, int msgno)
{
  HEADER *h = ctx->hdrs[msgno];
  MESSAGE *dest;

  int rc;
  short restore = 1;
  char oldpath[_POSIX_PATH_MAX];
  char newpath[_POSIX_PATH_MAX];
  char partpath[_POSIX_PATH_MAX];

  long old_body_offset = h->content->offset;
  long old_body_length = h->content->length;
  long old_hdr_lines = h->lines;

  if ((dest = mx_open_new_message (ctx, h, 0)) == NULL)
    return -1;

  if ((rc = mutt_copy_message (dest->fp, ctx, h,
                               M_CM_UPDATE, CH_UPDATE | CH_UPDATE_LEN)) == 0)
  {
    snprintf (oldpath, _POSIX_PATH_MAX, "%s/%s", ctx->path, h->path);
    strfcpy (partpath, h->path, _POSIX_PATH_MAX);

    if (ctx->magic == M_MAILDIR)
      rc = maildir_commit_message (dest, ctx, h);
    else
      rc = _mh_commit_message (dest, ctx, h, 0);

    mx_close_message (&dest);

    if (rc == 0) {
      unlink (oldpath);
      restore = 0;
    }

    /* 
     * Try to move the new message to the old place.
     * (MH only.)
     *
     * This is important when we are just updating flags.
     *
     * Note that there is a race condition against programs which
     * use the first free slot instead of the maximum message
     * number.  Mutt does _not_ behave like this.
     * 
     * Anyway, if this fails, the message is in the folder, so
     * all what happens is that a concurrently runnung mutt will
     * lose flag modifications.
     */

    if (ctx->magic == M_MH && rc == 0) {
      snprintf (newpath, _POSIX_PATH_MAX, "%s/%s", ctx->path, h->path);
      if ((rc = safe_rename (newpath, oldpath)) == 0)
        str_replace (&h->path, partpath);
    }
  }
  else
    mx_close_message (&dest);

  if (rc == -1 && restore) {
    h->content->offset = old_body_offset;
    h->content->length = old_body_length;
    h->lines = old_hdr_lines;
  }

  mutt_free_body (&h->content->parts);
  return rc;
}

static int mh_sync_message (CONTEXT * ctx, int msgno)
{
  HEADER *h = ctx->hdrs[msgno];

  if (h->attach_del || 
      (h->env && (h->env->refs_changed || h->env->irt_changed)))
    if (mh_rewrite_message (ctx, msgno) != 0)
      return -1;

  return 0;
}

static int maildir_sync_message (CONTEXT * ctx, int msgno)
{
  HEADER *h = ctx->hdrs[msgno];

  if (h->attach_del || 
      (h->env && (h->env->refs_changed || h->env->irt_changed))) {
    /* when doing attachment deletion/rethreading, fall back to the MH case. */
    if (mh_rewrite_message (ctx, msgno) != 0)
      return (-1);
  }
  else {
    /* we just have to rename the file. */

    char newpath[_POSIX_PATH_MAX];
    char partpath[_POSIX_PATH_MAX];
    char fullpath[_POSIX_PATH_MAX];
    char oldpath[_POSIX_PATH_MAX];
    char suffix[16];
    char *p;

    if ((p = strrchr (h->path, '/')) == NULL) {
      debug_print (1, ("%s: unable to find subdir!\n", h->path));
      return (-1);
    }
    p++;
    strfcpy (newpath, p, sizeof (newpath));

    /* kill the previous flags */
    if ((p = strchr (newpath, ':')) != NULL)
      *p = 0;

    maildir_flags (suffix, sizeof (suffix), h);

    snprintf (partpath, sizeof (partpath), "%s/%s%s",
              (h->read || h->old) ? "cur" : "new", newpath, suffix);
    snprintf (fullpath, sizeof (fullpath), "%s/%s", ctx->path, partpath);
    snprintf (oldpath, sizeof (oldpath), "%s/%s", ctx->path, h->path);

    if (str_cmp (fullpath, oldpath) == 0) {
      /* message hasn't really changed */
      return 0;
    }

    /* record that the message is possibly marked as trashed on disk */
    h->trash = h->deleted;

    if (rename (oldpath, fullpath) != 0) {
      mutt_perror ("rename");
      return (-1);
    }
    str_replace (&h->path, partpath);
  }
  return (0);
}

static int mh_sync_mailbox (CONTEXT * ctx, int unused, int *index_hint)
{
  char path[_POSIX_PATH_MAX], tmp[_POSIX_PATH_MAX];
  int i, j;

#if USE_HCACHE
  void *hc = NULL;
#endif /* USE_HCACHE */

  if (ctx->magic == M_MH)
    i = mh_check_mailbox (ctx, index_hint, 0);
  else
    i = maildir_check_mailbox (ctx, index_hint, 0);

  if (i != 0)
    return i;

#if USE_HCACHE
  if (ctx->magic == M_MAILDIR)
    hc = mutt_hcache_open (HeaderCache, ctx->path);
#endif /* USE_HCACHE */

  for (i = 0; i < ctx->msgcount; i++) {
    if (ctx->hdrs[i]->deleted
        && (ctx->magic != M_MAILDIR || !option (OPTMAILDIRTRASH))) {
      snprintf (path, sizeof (path), "%s/%s", ctx->path, ctx->hdrs[i]->path);
      if (ctx->magic == M_MAILDIR
          || (option (OPTMHPURGE) && ctx->magic == M_MH)) {
#if USE_HCACHE
        if (ctx->magic == M_MAILDIR)
          mutt_hcache_delete (hc, ctx->hdrs[i]->path + 3,
                              &maildir_hcache_keylen);
#endif /* USE_HCACHE */
        unlink (path);
      }
      else if (ctx->magic == M_MH) {
        /* MH just moves files out of the way when you delete them */
        if (*ctx->hdrs[i]->path != ',') {
          snprintf (tmp, sizeof (tmp), "%s/,%s", ctx->path,
                    ctx->hdrs[i]->path);
          unlink (tmp);
          rename (path, tmp);
        }

      }
    }
    else if (ctx->hdrs[i]->changed || ctx->hdrs[i]->attach_del ||
             (ctx->magic == M_MAILDIR
              && (option (OPTMAILDIRTRASH) || ctx->hdrs[i]->trash)
              && (ctx->hdrs[i]->deleted != ctx->hdrs[i]->trash))) {
      if (ctx->magic == M_MAILDIR) {
        if (maildir_sync_message (ctx, i) == -1)
          goto err;
      }
      else {
        if (mh_sync_message (ctx, i) == -1)
          goto err;
      }
    }
  }

#if USE_HCACHE
  if (ctx->magic == M_MAILDIR)
    mutt_hcache_close (hc);
#endif /* USE_HCACHE */

  if (ctx->magic == M_MH)
    mh_update_sequences (ctx);

  /* XXX race condition? */

  maildir_update_mtime (ctx);

  /* adjust indices */

  if (ctx->deleted) {
    for (i = 0, j = 0; i < ctx->msgcount; i++) {
      if (!ctx->hdrs[i]->deleted
          || (ctx->magic == M_MAILDIR && option (OPTMAILDIRTRASH)))
        ctx->hdrs[i]->index = j++;
    }
  }

  return 0;

err:
#if USE_HCACHE
  if (ctx->magic == M_MAILDIR)
    mutt_hcache_close (hc);
#endif /* USE_HCACHE */
  return -1;
}

static char *maildir_canon_filename (char *dest, const char *src, size_t l)
{
  char *t, *u;

  if ((t = strrchr (src, '/')))
    src = t + 1;

  strfcpy (dest, src, l);
  if ((u = strrchr (dest, ':')))
    *u = '\0';

  return dest;
}

static void maildir_update_tables (CONTEXT * ctx, int *index_hint)
{
  short old_sort;
  int old_count;
  int i, j;

  if (Sort != SORT_ORDER) {
    old_sort = Sort;
    Sort = SORT_ORDER;
    mutt_sort_headers (ctx, 1);
    Sort = old_sort;
  }

  old_count = ctx->msgcount;
  for (i = 0, j = 0; i < old_count; i++) {
    if (ctx->hdrs[i]->active && index_hint && *index_hint == i)
      *index_hint = j;

    if (ctx->hdrs[i]->active)
      ctx->hdrs[i]->index = j++;
  }

  mx_update_tables (ctx, 0);
  mutt_clear_threads (ctx);
}

static void maildir_update_flags (CONTEXT * ctx, HEADER * o, HEADER * n)
{
  /* save the global state here so we can reset it at the
   * end of list block if required.
   */
  int context_changed = ctx->changed;

  /* user didn't modify this message.  alter the flags to
   * match the current state on disk.  This may not actually
   * do anything, but we can't tell right now.  mutt_set_flag()
   * will just ignore the call if the status bits are
   * already properly set.
   */
  mutt_set_flag (ctx, o, M_FLAG, n->flagged);
  mutt_set_flag (ctx, o, M_REPLIED, n->replied);
  mutt_set_flag (ctx, o, M_READ, n->read);
  mutt_set_flag (ctx, o, M_OLD, n->old);

  /* mutt_set_flag() will set this, but we don't need to
   * sync the changes we made because we just updated the
   * context to match the current on-disk state of the
   * message.
   */
  o->changed = 0;

  /* if the mailbox was not modified before we made these
   * changes, unset the changed flag since nothing needs to
   * be synchronized.
   */
  if (!context_changed)
    ctx->changed = 0;
}


/* This function handles arrival of new mail and reopening of
 * maildir folders.  The basic idea here is we check to see if either
 * the new or cur subdirectories have changed, and if so, we scan them
 * for the list of files.  We check for newly added messages, and
 * then merge the flags messages we already knew about.  We don't treat
 * either subdirectory differently, as mail could be copied directly into
 * the cur directory from another agent.
 */
static int maildir_check_mailbox (CONTEXT * ctx, int *index_hint, int unused)
{
  struct stat st_new;           /* status of the "new" subdirectory */
  struct stat st_cur;           /* status of the "cur" subdirectory */
  char buf[_POSIX_PATH_MAX];
  int changed = 0;              /* bitmask representing which subdirectories
                                   have changed.  0x1 = new, 0x2 = cur */
  int occult = 0;               /* messages were removed from the mailbox */
  int have_new = 0;             /* messages were added to the mailbox */
  struct maildir *md;           /* list of messages in the mailbox */
  struct maildir **last, *p;
  int i;
  HASH *fnames;                 /* hash table for quickly looking up the base filename
                                   for a maildir message */

  if (!option (OPTCHECKNEW))
    return 0;

  snprintf (buf, sizeof (buf), "%s/new", ctx->path);
  if (stat (buf, &st_new) == -1)
    return -1;

  snprintf (buf, sizeof (buf), "%s/cur", ctx->path);
  if (stat (buf, &st_cur) == -1)
    return -1;

  /* determine which subdirectories need to be scanned */
  if (st_new.st_mtime > ctx->mtime)
    changed = 1;
  if (st_cur.st_mtime > ctx->mtime_cur)
    changed |= 2;

  if (!changed)
    return 0;                   /* nothing to do */

  /* update the modification times on the mailbox */
  ctx->mtime_cur = st_cur.st_mtime;
  ctx->mtime = st_new.st_mtime;

  /* do a fast scan of just the filenames in
   * the subdirectories that have changed.
   */
  md = NULL;
  last = &md;
  if (changed & 1)
    maildir_parse_dir (ctx, &last, "new", NULL);
  if (changed & 2)
    maildir_parse_dir (ctx, &last, "cur", NULL);

  /* we create a hash table keyed off the canonical (sans flags) filename
   * of each message we scanned.  This is used in the loop over the
   * existing messages below to do some correlation.
   */
  fnames = hash_create (1031);

  for (p = md; p; p = p->next) {
    maildir_canon_filename (buf, p->h->path, sizeof (buf));
    p->canon_fname = str_dup (buf);
    hash_insert (fnames, p->canon_fname, p, 0);
  }

  /* check for modifications and adjust flags */
  for (i = 0; i < ctx->msgcount; i++) {
    ctx->hdrs[i]->active = 0;
    maildir_canon_filename (buf, ctx->hdrs[i]->path, sizeof (buf));
    p = hash_find (fnames, buf);
    if (p && p->h) {
      /* message already exists, merge flags */
      ctx->hdrs[i]->active = 1;

      /* check to see if the message has moved to a different
       * subdirectory.  If so, update the associated filename.
       */
      if (str_cmp (ctx->hdrs[i]->path, p->h->path))
        str_replace (&ctx->hdrs[i]->path, p->h->path);

      /* if the user hasn't modified the flags on this message, update
       * the flags we just detected.
       */
      if (!ctx->hdrs[i]->changed)
        maildir_update_flags (ctx, ctx->hdrs[i], p->h);

      if (ctx->hdrs[i]->deleted == ctx->hdrs[i]->trash)
        ctx->hdrs[i]->deleted = p->h->deleted;
      ctx->hdrs[i]->trash = p->h->trash;

      /* this is a duplicate of an existing header, so remove it */
      mutt_free_header (&p->h);
    }
    /* This message was not in the list of messages we just scanned.
     * Check to see if we have enough information to know if the
     * message has disappeared out from underneath us.
     */
    else if (((changed & 1) && (!strncmp (ctx->hdrs[i]->path, "new/", 4))) ||
             ((changed & 2) && (!strncmp (ctx->hdrs[i]->path, "cur/", 4)))) {
      /* This message disappeared, so we need to simulate a "reopen"
       * event.  We know it disappeared because we just scanned the
       * subdirectory it used to reside in.
       */
      occult = 1;
    }
    else {
      /* This message resides in a subdirectory which was not
       * modified, so we assume that it is still present and
       * unchanged.
       */
      ctx->hdrs[i]->active = 1;
    }
  }

  /* destroy the file name hash */
  hash_destroy (&fnames, NULL);

  /* If we didn't just get new mail, update the tables. */
  if (occult)
    maildir_update_tables (ctx, index_hint);

  /* do any delayed parsing we need to do. */
  maildir_delayed_parsing (ctx, md);

  /* Incorporate new messages */
  have_new = maildir_move_to_context (ctx, &md);

  return occult ? M_REOPENED : (have_new ? M_NEW_MAIL : 0);
}

/* 
 * This function handles arrival of new mail and reopening of
 * mh/maildir folders. Things are getting rather complex because we
 * don't have a well-defined "mailbox order", so the tricks from
 * mbox.c and mx.c won't work here.
 *
 * Don't change this code unless you _really_ understand what
 * happens.
 *
 */

static int mh_check_mailbox (CONTEXT * ctx, int *index_hint, int unused)
{
  char buf[_POSIX_PATH_MAX];
  struct stat st, st_cur;
  short modified = 0, have_new = 0, occult = 0;
  struct maildir *md, *p;
  struct maildir **last = NULL;
  struct mh_sequences mhs;
  HASH *fnames;
  int i;

  if (!option (OPTCHECKNEW))
    return 0;

  strfcpy (buf, ctx->path, sizeof (buf));
  if (stat (buf, &st) == -1)
    return -1;

  /* create .mh_sequences when there isn't one. */
  snprintf (buf, sizeof (buf), "%s/.mh_sequences", ctx->path);
  if ((i = stat (buf, &st_cur) == -1) && errno == ENOENT) {
    char *tmp;
    FILE *fp = NULL;

    if (mh_mkstemp (ctx, &fp, &tmp) == 0) {
      safe_fclose (&fp);
      if (safe_rename (tmp, buf) == -1)
        unlink (tmp);
      mem_free (&tmp);
    }
  }

  if (i == -1 && stat (buf, &st_cur) == -1)
    modified = 1;

  if (st.st_mtime > ctx->mtime || st_cur.st_mtime > ctx->mtime_cur)
    modified = 1;

  if (!modified)
    return 0;

  ctx->mtime_cur = st_cur.st_mtime;
  ctx->mtime = st.st_mtime;

  memset (&mhs, 0, sizeof (mhs));

  md = NULL;
  last = &md;
  maildir_parse_dir (ctx, &last, NULL, NULL);
  mh_read_sequences (&mhs, ctx->path);
  mh_update_maildir (md, &mhs);
  mhs_free_sequences (&mhs);

  /* check for modifications and adjust flags */
  fnames = hash_create (1031);

  for (p = md; p; p = p->next)
    hash_insert (fnames, p->h->path, p, 0);

  for (i = 0; i < ctx->msgcount; i++) {
    ctx->hdrs[i]->active = 0;

    if ((p = hash_find (fnames, ctx->hdrs[i]->path)) && p->h &&
        (mutt_cmp_header (ctx->hdrs[i], p->h))) {
      ctx->hdrs[i]->active = 1;
      /* found the right message */
      if (!ctx->hdrs[i]->changed)
        maildir_update_flags (ctx, ctx->hdrs[i], p->h);

      mutt_free_header (&p->h);
    }
    else                        /* message has disappeared */
      occult = 1;
  }

  /* destroy the file name hash */

  hash_destroy (&fnames, NULL);

  /* If we didn't just get new mail, update the tables. */
  if (occult)
    maildir_update_tables (ctx, index_hint);

  /* Incorporate new messages */
  have_new = maildir_move_to_context (ctx, &md);

  return occult ? M_REOPENED : (have_new ? M_NEW_MAIL : 0);
}




/*
 * These functions try to find a message in a maildir folder when it
 * has moved under our feet.  Note that this code is rather expensive, but
 * then again, it's called rarely.
 */

FILE *_maildir_open_find_message (const char *folder, const char *unique,
                                  const char *subfolder)
{
  char dir[_POSIX_PATH_MAX];
  char tunique[_POSIX_PATH_MAX];
  char fname[_POSIX_PATH_MAX];

  DIR *dp;
  struct dirent *de;

  FILE *fp = NULL;
  int oe = ENOENT;

  snprintf (dir, sizeof (dir), "%s/%s", folder, subfolder);

  if ((dp = opendir (dir)) == NULL) {
    errno = ENOENT;
    return NULL;
  }

  while ((de = readdir (dp))) {
    maildir_canon_filename (tunique, de->d_name, sizeof (tunique));

    if (!str_cmp (tunique, unique)) {
      snprintf (fname, sizeof (fname), "%s/%s/%s", folder, subfolder,
                de->d_name);
      fp = fopen (fname, "r");  /* __FOPEN_CHECKED__ */
      oe = errno;
      break;
    }
  }

  closedir (dp);

  errno = oe;
  return fp;
}

FILE *maildir_open_find_message (const char *folder, const char *msg)
{
  char unique[_POSIX_PATH_MAX];
  FILE *fp;

  static unsigned int new_hits = 0, cur_hits = 0;       /* simple dynamic optimization */

  maildir_canon_filename (unique, msg, sizeof (unique));

  if ((fp =
       _maildir_open_find_message (folder, unique,
                                   new_hits > cur_hits ? "new" : "cur"))
      || errno != ENOENT) {
    if (new_hits < UINT_MAX && cur_hits < UINT_MAX) {
      new_hits += (new_hits > cur_hits ? 1 : 0);
      cur_hits += (new_hits > cur_hits ? 0 : 1);
    }

    return fp;
  }
  if ((fp =
       _maildir_open_find_message (folder, unique,
                                   new_hits > cur_hits ? "cur" : "new"))
      || errno != ENOENT) {
    if (new_hits < UINT_MAX && cur_hits < UINT_MAX) {
      new_hits += (new_hits > cur_hits ? 0 : 1);
      cur_hits += (new_hits > cur_hits ? 1 : 0);
    }

    return fp;
  }

  return NULL;
}


/*
 * Returns:
 * 1 if there are no messages in the mailbox
 * 0 if there are messages in the mailbox
 * -1 on error
 */
static int maildir_check_empty (const char *path)
{
  DIR *dp;
  struct dirent *de;
  int r = 1;                    /* assume empty until we find a message */
  char realpath[_POSIX_PATH_MAX];
  int iter = 0;

  /* Strategy here is to look for any file not beginning with a period */

  do {
    /* we do "cur" on the first iteration since its more likely that we'll
     * find old messages without having to scan both subdirs
     */
    snprintf (realpath, sizeof (realpath), "%s/%s", path,
              iter == 0 ? "cur" : "new");
    if ((dp = opendir (realpath)) == NULL)
      return -1;
    while ((de = readdir (dp))) {
      if (*de->d_name != '.') {
        r = 0;
        break;
      }
    }
    closedir (dp);
    iter++;
  } while (r && iter < 2);

  return r;
}

/*
 * Returns:
 * 1 if there are no messages in the mailbox
 * 0 if there are messages in the mailbox
 * -1 on error
 */
int mh_check_empty (const char *path)
{
  DIR *dp;
  struct dirent *de;
  int r = 1;                    /* assume empty until we find a message */

  if ((dp = opendir (path)) == NULL)
    return -1;
  while ((de = readdir (dp))) {
    if (mh_valid_message (de->d_name)) {
      r = 0;
      break;
    }
  }
  closedir (dp);

  return r;
}

static int mh_is_magic (const char* path, struct stat* st) {
  char tmp[_POSIX_PATH_MAX];

  if (S_ISDIR (st->st_mode)) {
    snprintf (tmp, sizeof (tmp), "%s/.mh_sequences", path);
    if (access (tmp, F_OK) == 0)
      return (M_MH);

    snprintf (tmp, sizeof (tmp), "%s/.xmhcache", path);
    if (access (tmp, F_OK) == 0)
      return (M_MH);

    snprintf (tmp, sizeof (tmp), "%s/.mew_cache", path);
    if (access (tmp, F_OK) == 0)
      return (M_MH);

    snprintf (tmp, sizeof (tmp), "%s/.mew-cache", path);
    if (access (tmp, F_OK) == 0)
      return (M_MH);

    snprintf (tmp, sizeof (tmp), "%s/.sylpheed_cache", path);
    if (access (tmp, F_OK) == 0)
      return (M_MH);

    /* 
     * ok, this isn't an mh folder, but mh mode can be used to read
     * Usenet news from the spool. ;-) 
     */

    snprintf (tmp, sizeof (tmp), "%s/.overview", path);
    if (access (tmp, F_OK) == 0)
      return (M_MH);
  }
  return (-1);
}

static int maildir_is_magic (const char* path, struct stat* st) {
  struct stat sb;
  char tmp[_POSIX_PATH_MAX];

  if (S_ISDIR (st->st_mode)) {
    snprintf (tmp, sizeof (tmp), "%s/cur", path);
    if (stat (tmp, &sb) == 0 && S_ISDIR (sb.st_mode))
      return (M_MAILDIR);
  }
  return (-1);
}

/* routines common to maildir and mh */
static mx_t* reg_mx (void) {
  mx_t* fmt = mem_calloc (1, sizeof (mx_t));
  fmt->local = 1;
  fmt->mx_access = access;
  fmt->mx_sync_mailbox = mh_sync_mailbox;
  return (fmt);
}

static int mh_commit (MESSAGE* msg, CONTEXT* ctx) {
  return (mh_commit_message (msg, ctx, NULL));
}

static int maildir_commit (MESSAGE* msg, CONTEXT* ctx) {
  return (maildir_commit_message (msg, ctx, NULL));
}

mx_t* mh_reg_mx (void) {
  mx_t* fmt = reg_mx ();
  fmt->type = M_MH;
  fmt->mx_check_empty = mh_check_empty;
  fmt->mx_is_magic = mh_is_magic;
  fmt->mx_open_mailbox = mh_read_dir;
  fmt->mx_open_new_message = mh_open_new_message;
  fmt->mx_check_mailbox = mh_check_mailbox;
  fmt->mx_commit_message = mh_commit;
  return (fmt);
}

mx_t* maildir_reg_mx (void) {
  mx_t* fmt = reg_mx ();
  fmt->type = M_MAILDIR;
  fmt->mx_check_empty = maildir_check_empty;
  fmt->mx_is_magic = maildir_is_magic;
  fmt->mx_open_mailbox = maildir_read_dir;
  fmt->mx_open_new_message = maildir_open_new_message;
  fmt->mx_check_mailbox = maildir_check_mailbox;
  fmt->mx_commit_message = maildir_commit;
  return (fmt);
}
